"""
The wntr.scenario package contains methods to define disaster scenarios and 
fragility/survival curves.
"""
from wntr.scenario.earthquake import Earthquake
from wntr.scenario.fragility_curve import FragilityCurve